% Logsweep1rt.m *** Logsweep to get TF and h(t), compute Reverberation Time
% has keyboard input for reverb filter frequency, mouse-click to get RT
% the left channel (ref) can be looped to determine the record latency,
% while the right channel (dut) measures the system.
% for most sound cards, the player is delayed wrt the recorder.
% the reference, if not detected, will be the sweep file (sweep).
%
% sweep is padded with zeros, giving 1/4 length clearance at sweep end. 
% the excitation sweep is windowed at both LF and HF (windosweep).
% TF is obtained from DUT/SWEEP, using unwindowed sweep.
% this allows sensible behaviour at LF and HF.
% TF2 is calculated from DUT/REF, nicely normalized but noisy near Nyquist.
% John Vanderkooy, March 2017, Audio Research Group, University of Waterloo
%% 0-----------------preliminaries and settings---------------------------%
clear all; clc; close all;
try
  pkg load signal;% for Octave signal processing toolbox
catch
end
disp('----------------start of program--------------------')
set(0,'DefaultLineLinewidth',1.5)
set(0,'DefaultAxesFontSize',12)
set(0,'DefaultAxesFontWeight','bold')
set(0,'DefaultAxesLineWidth',1.5)
Po=2e-5;% SPL reference pressure
c=343;% speed of sound
dBclearance=10;% white space above TF plots
dBspan=80;% total scale of TF plots
%----------------enter important parameters-----------------
sig_frac=0.50;% fraction of full scale
fs=44100;% must ensure that Windows settings are the same!
N=2^17;% make this larger if there is insufficient time clearance
S_dac=1.0;% enter value for your soundcard
S_adc=1.0;% enter value for your soundcard
% S_dac=-1.22;% UCA202
% S_adc=-1.40;% UCA202 (gain=3 Windows 7). This avoids ADC overload
% S_dac=0.98;% UA-1ex
% S_adc=0.5;% UA-1ex 
% S_dac=2.17*sqrt(2);% UA-25 May have internal +/-5V supply?
% S_adc=+6.0;% UA-25 also high.  
% S_dac=+1.59;% Focusrite 2i2 No, it does not inverts its DAC as article says
% S_adc=+1.16;% Focusrite 2i2 line input gain @ 12:00 o'clock
% S_dac=+1.148;% USB Dual Pre peak volts out for digital Full Scale 
% S_adc=+1.49;% USB Dual Pre JV pot minimum (gain=3, Windows 7)
%---------------system calibration factor CF for dut--------------------
system_type=0;% 1:acoustic, 0:electronic, -1:level-dependent,
power_amp_gain=10;% V/V acoustic
mic_cal=0.012;% V/Pa acoustic
mic_preamp_gain=10.0;% V/V acoustic
Vw=1;%2.83;% V/watt8ohms. acoustic Leave this=1 if not wanted
electronic_gain=1;% total gain in series with electronic system
if system_type==1 % acoustic
  CF=Vw/(power_amp_gain*mic_cal*mic_preamp_gain*Po);% SPL calibration factor
elseif system_type==0 % electronic system
  CF=1/electronic_gain;% electronic
else
  CF=1/(sig_frac*mic_cal*mic_preamp_gain*Po);% SPL level-dependent
end
%-------------system parameters ----------------------------
Npad=N/4;% this is the total zeropad, added to end of play file
Ns=N-Npad;% most of array is used for sweep
disp(['fs: ' num2str(fs) '  N: ' num2str(N) '  duration: ' num2str(N/fs)])
disp(['signal fraction: ' num2str(sig_frac) '  time clearance: ' num2str(N/(4*fs))])
t=linspace(0,(N-1)/fs,N)';% column vector
ts=linspace(0,(Ns-1)/fs,Ns)';% to calculate sweep

%% 1-----------------------calculate logsweep-----------------------------%
f_start=5;% beginning of turnon half-Hann
f1=10;% end of turnon half-hann
f2=0.91*fs/2;% beginning of turnoff half-Hann
f_stop=fs/2;% end of turnoff half-Hann
Ts=Ns/fs;% sweep duration.  This is N-Npad samples in length.
Ls=Ts/log(f_stop/f_start);% time for frequency to increase by factor e
sweep=zeros(N,1);% initialize
sweep(1:Ns)=sin(2*pi*f_start*Ls*(exp(ts/Ls)-1));% logsweep
%------------------tapered sweep window------------------------------
indexf1=round(fs*Ls*log(f1/f_start))+1;% end of starting taper
indexf2=round(fs*Ls*log(f2/f_start))+1;% beginning of ending taper
windo=ones(N,1);
windo(1:indexf1)=0.5*(1-cos(pi*(1:indexf1)/indexf1));% pre-taper
windo(indexf2+1:Ns)=0.5*(1+cos(pi*(1:Ns-indexf2)/(Ns-indexf2)))';% post-taper
windo(Ns+1:N)=0; % zeropad end of sweep
windosweep=windo.*sweep;% tapered at each end for output to DAC

figure(10)
plot(t,sweep,'b')
grid on;hold on
plot(t,windosweep,'r')
xlabel('time[s]')
axis([0 N/fs -1.5 1.5])
legend('raw sweep','windowed sweep')
title('Sweep')
disp(['f_start*Ls: ',num2str(f_start*Ls) '  Ls: ' num2str(Ls)])
disp('finished sweep generation...')
%% 2------------data gathering: send out sweep, record system output------%
y=sig_frac*[windosweep -windosweep];% antiphase avoids codec midtap modulation
rec = audiorecorder(fs,16,2);
ply = audioplayer(y,fs,16);
play(ply); %this should play and allow immediate further execution
recordblocking(rec,N/fs);% program waits until all samples recorded
z = getaudiodata(rec,'double');% this retrieves the recorded data
disp('finished recording...');

disp('Some sound cards act strangely. Check carefully!')
if max(abs(z(:,1)))>0.999
   disp('Check for ch0 REF record saturation!');
end
if max(abs(z(:,2)))>0.999
   disp('Check for ch1 DUT record saturation!');
end
%----------------------- end of data gathering ---------------------
%load('Logsweep1datart.mat');
ref=z(1:N,1);%if we use left channel as reference
dut=z(1:N,2);
N=length(dut);
%save('Logsweep1data.mat','sweep','z','N','fs','sig_frac','CF');
clear y z;

disp(['refrmsLSBs: ' num2str(sqrt(2^30*sum(ref.^2)/N))])
disp(['dutrmsLSBs: ' num2str(sqrt(2^30*sum(dut.^2)/N))])

figure(20);
plot(t,ref,'b');
grid on;hold on
plot(t,dut,'r');
axis([0 N/fs -1.5 1.5])
legend('raw ref','raw dut');
xlabel('Time [s]')
title('Recorded responses');

%% 3----------determine record/play delay using crosscorrelation----------%
lags=N/2;% large enough to catch most delays
if max(ref) < 0.1*max(dut) % automatic reference selection
    X=xcorr(sweep,dut,lags);% in case reference is low, use data itself
else
    X=xcorr(sweep,ref,lags);% this uses recorded reference
end
[~,nmax]=max(abs(X));

figure(30)
tl=t-lags/fs;% centred plot
plot(tl(1:N),X(1:N))
grid on
legend('crosscorrelation')
xlabel('time [s]')
title('recorder leads player-----|------recorder lags player')

offset=nmax-lags-1;
disp(['record offset: ' num2str(offset) '  time: ' num2str(offset/fs)])
if abs(offset) > Npad 
    disp('******INSUFFICIENT TIME CLEARANCE!******')
    disp('******INSUFFICIENT TIME CLEARANCE!******')
    disp('******INSUFFICIENT TIME CLEARANCE!******')
end
disp('negative offset means player lags recorder!')

lwindo=ones(N,1);
lwindo(1:indexf1)=0.5*(1-cos(pi*(1:indexf1)/indexf1));% LF pre-taper
lwindosweep=lwindo.*sweep;
% remove play-record delay by shifting computer sweep array
%sweep=circshift(sweep,-offset);
lwindosweep=circshift(lwindosweep,-offset);
%% 4-----------calculate TFs using Frequency Domain ratios----------------%
% all frequency variables are meant to be voltage spectra
%SWEEP=sig_frac*S_dac*fft(sweep);
LWINDOSWEEP=sig_frac*S_dac*fft(lwindosweep);
REF=S_adc*fft(ref);
DUT=CF*S_adc*fft(dut);% calibration factor normally has dimensions volts

TF=DUT./LWINDOSWEEP;% this has good Nyquist behaviour
%TF=DUT./SWEEP;
TF2=DUT./REF;
ft=((1:N/2+1)'-1)*fs/N;
ft(1)=NaN;% this avoids non-positive data in semilogx warning from Octave

top=20*log10(max(abs(TF(2:N/2+1))));% highest dB in plot
figure(100);
semilogx(ft,20*log10(abs(TF(1:N/2+1))),'b');
grid on;
axis([fs/N fs/2 top+dBclearance-dBspan top+dBclearance])
legend('DUT/SWP')
xlabel('frequency [Hz]');ylabel('dB')
title('Transfer Function');

top=20*log10(max(abs(TF2(2:N/2+1))));% highest dB in plot
figure(110);
semilogx(ft,20*log10(abs(TF2(1:N/2+1))),'b');
grid on;
axis([fs/N fs/2 top+dBclearance-dBspan top+dBclearance])
legend('DUT/REF')
xlabel('frequency [Hz]');ylabel('dB')
title('Transfer Function 2');

%% 5---------------load previous file and compare TF ---------------------%
try
    load('old_TF.mat')% loads old_TF
    % this figure will appear when file exists
    top=20*log10(max(abs(TF(2:N/2+1))));% highest dB in plot
    figure(120);
    semilogx(ft,20*log10(abs(TF(1:N/2+1))),'b');
    grid on;hold on
    if length(old_TF) == length(TF)
     semilogx(ft,20*log10(abs(old_TF(1:N/2+1))),'r');
    else
     disp('Different file lengths: do another run')   
    end %if length
    axis([1 fs/2 top+dBclearance-dBspan top+dBclearance])
    legend('new DUT/SWP','old DUT/SWP')
    xlabel('frequency [Hz]');ylabel('dB')
    title('old & new Transfer Functions');
catch
    disp('continuing...')
end %try
old_TF=TF;
save('old_TF.mat','old_TF');% saves current TF file
% all the programs are the same to this point--------------------
%% 6 ---------obtain impulse response ------------------------------------%
h=real(ifft(TF));% 'real' should not be neccesary
h=circshift(h,round(0.001*fs));% 1 ms shift to capture spike wraparound at t=0
nh5=round(fs*Ls*log(5));
h(N-nh5:N)=0;% removes harmonic impulse spikes
figure(300);
plot(t,h,'b')
grid on;
axis([0 1.05*N/fs min(h)-0.05*max(h) 1.2*max(h)])
xlabel('time [s]')
title('h(t) from DUT/SWP');
%--------------compute clarity C50 or C80-------------------
energy=h.^2;
index50=round(0.05*fs);%50 ms from t=0
early_energy=sum(energy(1:index50));
late_energy=sum(energy(index50+1:N));
C50=10*log(early_energy/late_energy);
disp(['C50 = ' num2str(C50)])
%% 7 ----------------plot TF magnitude and phases-------------------------%
% plot first half of TF. 2nd half is repeat conjugate even, since h is real
TF=fft(h);
oct_width=0.33;% power-preserving relative frequency smoothing for TF
TFp=pwroctsmooth(TF,oct_width);

top=20*log10(max(abs(TF(2:N/2+1))));% highest dB in plot
figure(400);
semilogx(ft,20*log10(abs(TF(1:N/2+1))),'b');
grid on;hold on;
semilogx(ft,20*log10(abs(TFp(1:N/2+1))),'r');
axis([fs/10000 fs/2 top+dBclearance-dBspan top+dBclearance])
legend('unsmoothed','smoothed')
xlabel('frequency [Hz]')
ylabel('dB')
title(['DUT/SWP ' num2str(oct_width) 'octave Smoothed Response']);

%% 8-----------calculate reverb decay for a chosen octave band------------%
% bandpass filter
disp(' ')
centre_freq=input('Octave-band centre frequency [Hz], then ENTER: ');
W1=2*centre_freq/(fs*sqrt(2));
W2=2*centre_freq*sqrt(2)/fs;
[b,a]=butter(2,[W1 W2]);% filitfilt makes it 4th order
h_filt=filtfilt(b,a,h);% linear phase filter, no causal time smear
decay_plot=flip(cumsum(flip(h_filt.^2)));% reverse integration
decay_plot=decay_plot+decay_plot(1)*1e-6; % limit decay to -60dB
decay_plot=decay_plot/decay_plot(1); % normalize to 0 dB
% -------------------position final reverb plot ------------------
set(0,'Units','normalized')
hr=figure(500);
p=get(hr,'Position');
tilenum=1;% move plot by 10% for each tilenumber increase
set(hr,'Position',[p(1)*(1+tilenum/10) p(2)*(1-tilenum/10) p(3) p(4)])
plot(t,10*log10(decay_plot),'b');
grid on;
axis([0 1.0*N/fs -70 10])% make N/fs smaller to expand the plot
xlabel('time [s]')
ylabel('dB')
title('Schroeder decay plot');

%GUI-------------calculate RT from decay plot----------------------------%
xl=xlim();yl=ylim();
x=xl(1)+0.05*(xl(2)-xl(1));
y=yl(1)+0.1*(yl(2)-yl(1));
text(x,y,'Select Decay: start, end','fontsize',14);
[X,Y]=ginput(2);
RT=(X(2)-X(1))*60/(Y(1)-Y(2));
disp(['RT = ' num2str(RT)])
disp('-------------------finished--------------------') 
